@@ -49,6 +49,65 @@
 
 #include "htp.h"
 
+/*
+ * SYNOPSIS:
+ *   static char *mime_encode(
+ *                            const unsigned char *data
+ *                           );
+ *
+ * ARGUMENTS:
+ *   const unsigned char *data  Data to MIME64 encode.
+ *
+ * RETURN VALUE:
+ *   This function returns a pointer to the MIME64 encoded data.
+ *   The return buffer has been allocated with malloc() and must be free()'d.
+ *
+ * NOTES:
+ *   Free this buffer after use.
+ *
+ */
+static char *mime_encode(const unsigned char *data) {
+	static char mimeabet[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+	unsigned long buffer;
+	size_t datalen, retlen, outlen;
+	char *ret;
+	int i, x=0;
+
+	datalen = strlen(data);
+
+	/* Size of actual charectars to fill. */
+	outlen = (int) (((double) datalen) * 1.3333333 + 0.9);
+
+	/* Size of buffer, atleast as large as "outlen" but probably larger. */
+	retlen = (int) ((((double) datalen) / 3.0 + 0.9) * 4);
+
+	ret = malloc(retlen + 1);
+
+	if (ret == NULL) {
+		return(NULL);
+	}
+
+	for (i = 0; i < datalen; i += 3) {
+		buffer = (data[i] << 16) | (data[i + 1] << 8) | data[i + 2];
+		ret[x++] = mimeabet[(buffer & 0xfc0000) >> 18];
+		if (x >= outlen) break;                                     
+		ret[x++] = mimeabet[(buffer & 0x3f000) >> 12];
+		if (x >= outlen) break;                                                                      
+		ret[x++] = mimeabet[(buffer & 0xfc0) >> 6];
+		if (x >= outlen) break;
+		ret[x++] = mimeabet[buffer & 0x3f];
+		if (x >= outlen) break;
+	}
+
+	while (x != retlen) {
+		ret[x++]='=';
+	}
+
+	ret[x] = '\0';
+
+	return(ret);
+}
+
 /*
  * SYNOPSIS:
  *   static time_t get_gmtoffset(void);
@@ -200,7 +259,7 @@ double set_clock(const struct timeval *tvdata, int allow_adj, int ignoreusec) {
 	}
 
 #ifdef HAVE_ADJTIME
-	if ((tvdelta.tv_sec < 60 && tvdelta.tv_sec > -60) && allow_adj) {
+	if ((tvdelta.tv_sec < 3 && tvdelta.tv_sec > -3) && allow_adj) {
 		chkret = adjtime(&tvdelta, NULL);
 		if (chkret >= 0) {
 			retval = tvdelta.tv_sec + (((double) tvdelta.tv_usec) / 1000000.0);
@@ -215,7 +274,7 @@ double set_clock(const struct timeval *tvdata, int allow_adj, int ignoreusec) {
 		chkret = settimeofday(&tvinfo, NULL);
 		if (chkret >= 0) {
 			retval = tvdelta.tv_sec + (((double) tvdelta.tv_usec) / 1000000.0);
-#ifdef HAVE_SYSLOOCCOC
+#ifdef HAVE_SYSLOG
 			syslog(LOG_NOTICE, "Adjusting clock by %f seconds (step).", retval);
 #endif
 		}
@@ -238,6 +297,7 @@ double set_clock(const struct timeval *tvdata, int allow_adj, int ignoreusec) {
  *   unsigned int port       Destination port
  *   const char *proxyhost   Proxy to send request through
  *   unsigned int proxyport  Proxy port to connect
+ *   const char *credentials Authentication credentials (NULL for no auth)
  *
  * RETURN VALUE:
  *   This function returns the GMT/UTC time_t value that is specified
@@ -250,7 +310,7 @@ double set_clock(const struct timeval *tvdata, int allow_adj, int ignoreusec) {
  *   directly (and proxyport is ignored.)
  *
  */
-static time_t get_server_time(const char *host, unsigned int port, const char *proxyhost, unsigned int proxyport) {
+static time_t get_server_time(const char *host, unsigned int port, const char *proxyhost, unsigned int proxyport, const char *credentials) {
 	unsigned int         server_s;		/* Server socket descriptor */
 	struct sockaddr_in   server_addr;	/* Server Internet address */
 	char                 out_buf[2048];	/* Output buffer for HEAD request */
@@ -294,8 +354,10 @@ static time_t get_server_time(const char *host, unsigned int port, const char *p
 	/* Send the most simple HEAD request */
 	if (proxyhost == NULL) {
 		snprintf(out_buf, sizeof(out_buf), "HEAD / HTTP/1.0\r\nPragma: no-cache\r\nCache-Control: Max-age=0\r\n\r\n");
-	} else {
+	} else if (credentials == NULL) {
 		snprintf(out_buf, sizeof(out_buf), "HEAD http://%s:%i/ HTTP/1.0\r\nPragma: no-cache\r\nCache-Control: Max-age=0\r\n\r\n", host, port);
+	} else {
+		snprintf(out_buf, sizeof(out_buf), "HEAD http://%s:%i/ HTTP/1.0\r\nPragma: no-cache\r\nCache-Control: Max-age=0\r\nProxy-Authorization: Basic %s\r\n\r\n", host, port, credentials);
 	}
 	send(server_s, out_buf, strlen(out_buf), 0);
 
@@ -441,6 +503,10 @@ static int compare_time(const void *a, const void *b) {
  *                                          array.
  *   const char *proxyhost                  Proxy hostname to connect through.
  *   unsigned int proxyport                 Port to connect to proxy on.
+ *   const char *username                   Username for authentication (NULL
+ *                                          for no authentication.)
+ *   const char *password                   Password for authentication (NULL
+ *                                          for no authentication.)
  *
  * RETURN VALUE:
  *   This function returns the best approximation of the current time in
@@ -458,7 +524,7 @@ static int compare_time(const void *a, const void *b) {
  *   to it you must store the values elsewhere.
  *
  */
-struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int totaltimeservers, const char *proxyhost, unsigned int proxyport) {
+struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int totaltimeservers, const char *proxyhost, unsigned int proxyport, const char *username, const char *password) {
 	unsigned int timeind = 0, meantimeind = 0;
 	unsigned int totaltimes = 0, totalstddevtimes = 0;
 	unsigned int port = 0;
@@ -472,17 +538,23 @@ struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int tot
 	struct timeval *tmptime;
 	long newusecs = 0;
 	char *host = NULL;
+	char *authcredentials = NULL, tmpbuf[128];
 	int stddevtimes = 0;
 
 	if (gettimeofday(&start_time, NULL) < 0) {
 		return(NULL);
 	}
 
+	if (username != NULL && password != NULL) {
+		snprintf(tmpbuf, sizeof(tmpbuf), "%s:%s", username, password);
+		authcredentials = mime_encode(tmpbuf);
+	}
+
 	for (timeind = 0; timeind < totaltimeservers; timeind++) {	
 		host = timeservers[timeind].host;
 		port = timeservers[timeind].port;
 
-		timeval = get_server_time(host, port, proxyhost, proxyport);
+		timeval = get_server_time(host, port, proxyhost, proxyport, authcredentials);
 		if (timeval < 0) {
 			continue;
 		}
@@ -502,6 +574,9 @@ struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int tot
 
 	/* Return failure if no time servers could be reached. */
 	if (totaltimes == 0) {
+		if (authcredentials != NULL) {
+			free(authcredentials);
+		}
 		return(NULL);
 	}
 
@@ -563,6 +638,9 @@ struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int tot
 	}
 
 	if (newtimeval <= 0) {
+		if (authcredentials != NULL) {
+			free(authcredentials);
+		}
 		return(NULL);
 	}
 
@@ -576,5 +654,10 @@ struct timeval *htp_calctime(struct timeserver_st *timeservers, unsigned int tot
 	}
 	retval.tv_usec = newusecs;
 
+	/* Free our buffers. */
+	if (authcredentials != NULL) {
+		free(authcredentials);
+	}
+
 	return(&retval);
 }
